#include "inc/hw_types.h"
#include "inc/hw_memmap.h"
#include "driverlib/sysctl.h"
#include "driverlib/interrupt.h"
#include "driverlib/systick.h"
#include "driverlib/gpio.h"
#include "grlib/grlib.h"
#include "drivers/kitronix320x240x16_ssd2119_8bit.h"
#include "drivers/set_pinout.h"
#include "drivers/thumbwheel.h"
#include "utils/ustdlib.h"

#define USER_BTN_PORT			GPIO_PORTJ_BASE
#define USER_BTN_PIN			GPIO_PIN_7

#define PADDLE_WIDTH			5
#define PADDLE_SIZE				20
#define PADDLE_GRACE			5
#define BALL_SIZE				4

#define INITIAL_DELAYPERIOD		15000
#define DELAY_SPEEDUP			250

static volatile unsigned int timeTicks = 0;
static volatile int newPaddleY = 0;
static volatile int delayPeriod;

void systickISR(void) {
	timeTicks++;
	if ((timeTicks & 127) == 0) {
		if (delayPeriod > 1000) {
			delayPeriod -= DELAY_SPEEDUP;
		}
	}
}

static void delay(int aLength) {
	do {
		__asm__ __volatile__("nop");
		aLength--;
	} while (aLength);
}

static void thumbwheelCallback(unsigned short aValue) {
	newPaddleY = ((3000 - aValue) * 240 / 3000);
}

static int buttonPressed(void) {
	return (GPIOPinRead(USER_BTN_PORT, USER_BTN_PIN) & USER_BTN_PIN) == 0;
}

static void drawPaddle(tContext *aGraphics, int aPaddleY, unsigned char aClear) {
	tRectangle rect;

	rect.sXMin = 0;
	rect.sXMax = 5;
	rect.sYMin = aPaddleY - 20;
	rect.sYMax = aPaddleY + 20;
	GrContextForegroundSet(aGraphics, aClear ? ClrDarkBlue: ClrWhite);
	GrRectFill(aGraphics, &rect);
}

static void gameOver(tContext *aGraphics) {
	GrContextForegroundSet(aGraphics, ClrRed);
	GrContextFontSet(aGraphics, &g_sFontCm40);
	GrStringDrawCentered(aGraphics, "Game Over", -1,
							GrContextDpyWidthGet(aGraphics) / 2,
							GrContextDpyHeightGet(aGraphics) / 2,
						 	0);

	char buffer[32];
	unsigned int t = timeTicks;
	usnprintf(buffer, sizeof(buffer), "You lasted %d.%02d secs", t / 100, t % 100);
	
	GrContextForegroundSet(aGraphics, ClrYellow);
	GrContextFontSet(aGraphics, &g_sFontCm14);
	GrStringDrawCentered(aGraphics, buffer, -1,
							GrContextDpyWidthGet(aGraphics) / 2,
							(GrContextDpyHeightGet(aGraphics) / 2) + 30,
						 	0);
	
	GrContextForegroundSet(aGraphics, ClrWhite);
	GrContextFontSet(aGraphics, &g_sFontCm14);
	GrStringDrawCentered(aGraphics, "Press button to try again...", -1,
							GrContextDpyWidthGet(aGraphics) / 2,
							(GrContextDpyHeightGet(aGraphics) / 2) + 50,
						 	0);

	while (!buttonPressed()) {
	}
}

static void playPong(tContext *aGraphics) {
	tRectangle rect;

	rect.sXMin = 0;
	rect.sYMin = 0;
	rect.sXMax = GrContextDpyWidthGet(aGraphics) - 1;
	rect.sYMax = GrContextDpyHeightGet(aGraphics) - 1;
	GrContextForegroundSet(aGraphics, ClrDarkBlue);
	GrRectFill(aGraphics, &rect);

	int paddleY = 30;
	newPaddleY = paddleY;
	int posX = 40;
	int posY = 40;
	int speedX = 1;
	int speedY = 1;
	delayPeriod = INITIAL_DELAYPERIOD;
	timeTicks = 0;
	while (1) {
		/* Delete old ball */
		GrContextForegroundSet(aGraphics, ClrDarkBlue);
		GrCircleDraw(aGraphics, posX, posY, BALL_SIZE);

		/* Delete old paddle */
		if (newPaddleY != paddleY) {
			drawPaddle(aGraphics, paddleY, 1);
		}

		/* Recalculate movement */
		posX += speedX;
		posY += speedY;
		if (posX > (GrContextDpyWidthGet(aGraphics) - 1 - BALL_SIZE)) {
			posX = (GrContextDpyWidthGet(aGraphics) - 1 - BALL_SIZE);
			speedX = -speedX;
		} else if (posX < BALL_SIZE + PADDLE_WIDTH + 1) {
			if ((posY >= paddleY - PADDLE_SIZE - PADDLE_GRACE) && (posY <= paddleY + PADDLE_SIZE + PADDLE_GRACE)) {
				posX = BALL_SIZE + PADDLE_WIDTH + 1;
				speedX = -speedX;
			} else {
				gameOver(aGraphics);
				return;
			}
		}

		if (posY > (GrContextDpyHeightGet(aGraphics) - 1 - BALL_SIZE)) {
			posY = (GrContextDpyHeightGet(aGraphics) - 1 - BALL_SIZE);
			speedY = -speedY;
		} else if (posY < BALL_SIZE) {
			posY = BALL_SIZE;
			speedY = -speedY;
		}

		GrContextForegroundSet(aGraphics, ClrWhite);
		GrCircleDraw(aGraphics, posX, posY, BALL_SIZE);
		
		if (newPaddleY != paddleY) {
			paddleY = newPaddleY;
			drawPaddle(aGraphics, paddleY, 0);
		}

		GrFlush(aGraphics);
		ThumbwheelTrigger();
		delay(delayPeriod);
	}
}

int main(void) {
	tContext sContext;

	// Set the clocking to run directly from the crystal.
	SysCtlClockSet(SYSCTL_SYSDIV_1 | SYSCTL_USE_OSC | SYSCTL_XTAL_16MHZ |
					SYSCTL_OSC_MAIN);

	// Initialize the device pinout appropriately for this board.
	PinoutSet();

	// Initialize the display driver.
	Kitronix320x240x16_SSD2119Init();

	// Initialize the graphics context.
	GrContextInit(&sContext, &g_sKitronix320x240x16_SSD2119);

	ThumbwheelInit();
	ThumbwheelCallbackSet(thumbwheelCallback);

	/* Enable Systick */
	SysTickPeriodSet(SysCtlClockGet() / 100);
	SysTickEnable();
	SysTickIntEnable();

	GPIODirModeSet(USER_BTN_PORT, USER_BTN_PIN, GPIO_DIR_MODE_IN);
	GPIOPadConfigSet(USER_BTN_PORT, USER_BTN_PIN, GPIO_STRENGTH_2MA,
								GPIO_PIN_TYPE_STD_WPU);

	// Enable Interrupts
	IntMasterEnable();

	while (1) {
		playPong(&sContext);
	}
}
